#include "../std.h"
#include <iostream>

//#include "../misc/condition.h"
//#include "../misc/checksum.h"
//#include "../misc/inttostring.h"
#include "../https/SRM/srm1_soapH.h"
extern SOAP_NMAC struct Namespace srm1_soap_namespaces[];
#include "../https/SRM/srm_url.h"
#include "../https/SRM/srm_client.h"
#include "../misc/random.h"
//#include "../misc/time_utils.h"
//#include "../../userinterface/CertInfo.h"
#include "../misc/log_time.h"
#include "databufferpar.h"
#include "datapoint.h"
#include "datahandle.h"

bool DataHandle::start_reading_srm(DataBufferPar &buf) {
  if(r_handle) return false; // must be operation in progress
  SRM_URL srm_url(c_url.c_str());
  if(!srm_url) return false;
  SRMClient client(srm_url.ContactURL().c_str());
  if(!client) return false;
  r_url=NULL; r_handle=NULL; srm_request=NULL;
  list<string> turls;
  if(!no_checks) {
    odlog(2)<<"start_reading_srm: looking for metadata: "<<c_url.c_str()<<endl;
    unsigned long long int size;
    string checksum;
    if(!client.info(srm_url.FileName().c_str(),srm_url,size,checksum)) return false;
    /* provide some metadata */
    odlog(1)<<"start_reading_srm: obtained size: "<<size<<endl;
    if(size > 0) url->meta_size(size);
    odlog(1)<<"start_reading_srm: obtained checksum: "<<checksum<<endl;
    if(checksum.length() > 0) url->meta_checksum(checksum.c_str());
  };
  srm_request = new SRMRequest(); if(!srm_request) goto error_exit;
  if(!client.getTURLs(*srm_request,srm_url.FileName().c_str(),srm_url,turls)) {
    goto error_exit;
  };
  client.disconnect();
  // Choose handled URL randomly
  for(;;) {
    if(turls.size() <= 0) break;
    int n = Random::get(turls.size()-1);
    list<string>::iterator i = turls.begin(); for(;n;++i,++n) { };
    if(i == turls.end()) continue;
    // Avoid redirection to SRM
    odlog(2)<<"Checking URL returned by SRM: "<<*i<<endl;
    if(strncasecmp(i->c_str(),"srm://",6) == 0) { turls.erase(i); continue; };
    // Try to use this TURL
    r_url = DataPoint::CreateInstance(i->c_str());
    // check if url can be handled
    if(!r_url) { turls.erase(i); continue; };
    if(r_url->meta()) { delete r_url; r_url=NULL; turls.erase(i); continue; };
    break;
  };
  if(r_url == NULL) {
    odlog(1)<<"SRM returned no useful Transfer URLs: "<<c_url<<endl;
    goto error_exit;
  };
  r_handle = new DataHandle(r_url);
  r_handle->ftp_threads=ftp_threads;
  r_handle->cacheable=cacheable;
  r_handle->linkable=linkable;
  r_handle->force_secure=force_secure;
  r_handle->force_passive=force_passive;
  odlog(1)<<"Redirecting to new URL: "<<*r_url<<endl;
  if(!(r_handle->start_reading(buf))) goto error_exit;
  return true;
error_exit:
  if(r_handle) delete r_handle; r_handle=NULL;
  if(r_url) delete r_url; r_url=NULL;
  if(srm_request) delete srm_request; srm_request=NULL;
  return false;
}

bool DataHandle::stop_reading_srm(void) {
  if(!r_handle) return true;
  bool r = r_handle->stop_reading();
  delete r_handle;
  if(r_url) delete r_url;
  if(srm_request) {
    SRM_URL srm_url(c_url.c_str());
    if(srm_url) {
      SRMClient client(srm_url.ContactURL().c_str());
      if(client) {
        client.release(*srm_request,srm_url.FileName().c_str(),srm_url);
      };
    };
    delete srm_request;
  };
  r_handle=NULL; r_url=NULL; srm_request=NULL;
  return r;
}

bool DataHandle::start_writing_srm(DataBufferPar &buf) {
  if(r_handle) return false; // must be operation in progress
  SRM_URL srm_url(c_url.c_str());
  if(!srm_url) return false;
  SRMClient client(srm_url.ContactURL().c_str());
  if(!client) return false;
  r_url=NULL; r_handle=NULL; srm_request=NULL;
  list<string> turls;
  srm_request = new SRMRequest(); if(!srm_request) goto error_exit;
  if(!client.putTURLs(*srm_request,srm_url.FileName().c_str(),srm_url,turls,url->meta_size())) goto error_exit;
  client.disconnect();
  // Choose handled URL randomly
  for(;;) {
    if(turls.size() <= 0) break;
    int n = Random::get(turls.size()-1);
    list<string>::iterator i = turls.begin(); for(;n;++i,++n) { };
    if(i == turls.end()) continue;
    odlog(2)<<"Checking URL returned by SRM: "<<*i<<endl;
    // Avoid redirection to SRM
    if(strncasecmp(i->c_str(),"srm://",6) == 0) { turls.erase(i); continue; };
    // Try to use this TURL
    r_url = DataPoint::CreateInstance(i->c_str());
    // check if url can be handled
    if(!r_url) { turls.erase(i); continue; };
    if(r_url->meta()) { delete r_url; r_url=NULL; turls.erase(i); continue; };
    break;
  };
  if(r_url == NULL) {
    odlog(1)<<"SRM returned no useful Transfer URLs: "<<c_url<<endl;
    goto error_exit;
  };
  r_handle = new DataHandle(r_url);
  r_handle->ftp_threads=ftp_threads;
  r_handle->cacheable=cacheable;
  r_handle->linkable=linkable;
  r_handle->force_secure=force_secure;
  r_handle->force_passive=force_passive;
  odlog(1)<<"Redirecting to new URL: "<<*r_url<<endl;
  if(!(r_handle->start_writing(buf))) goto error_exit;
  return true;
error_exit:
  if(r_handle) delete r_handle; r_handle=NULL;
  if(r_url) delete r_url; r_url=NULL;
  if(srm_request) delete srm_request; srm_request=NULL;
  return false;
}

bool DataHandle::stop_writing_srm(void) {
  if(!r_handle) return true;
  bool r = r_handle->stop_reading();
  delete r_handle;
  if(r_url) delete r_url;
  if(srm_request) {
    SRM_URL srm_url(c_url.c_str());
    if(srm_url) {
      SRMClient client(srm_url.ContactURL().c_str());
      if(client) {
        client.release(*srm_request,srm_url.FileName().c_str(),srm_url);
      };
    };
    delete srm_request;
  };
  r_handle=NULL; r_url=NULL; srm_request=NULL;
  return r;
}

bool DataHandle::remove_srm(void) {
  SRM_URL srm_url(c_url.c_str());
  if(!srm_url) return false;
  SRMClient client(srm_url.ContactURL().c_str());
  if(!client) return false;
  odlog(2)<<"remove_srm: deleting: "<<c_url.c_str()<<endl;
  if(!client.remove(srm_url.FileName().c_str(),srm_url)) return false;
  return true;
}

bool DataHandle::list_files_srm(list<DataPoint::FileInfo> &files,bool resolve) {
  SRM_URL srm_url(c_url.c_str());
  if(!srm_url) return false;
  SRMClient client(srm_url.ContactURL().c_str());
  if(!client) return false;
  odlog(2)<<"list_files_srm: looking for metadata: "<<c_url.c_str()<<endl;
  unsigned long long int size;
  string checksum;
  if(!client.info(srm_url.FileName().c_str(),srm_url,size,checksum)) return false;
  if(size > 0) url->meta_size(size);
  if(checksum.length() > 0) url->meta_checksum(checksum.c_str());
  list<DataPoint::FileInfo>::iterator f =
     files.insert(files.end(),DataPoint::FileInfo(srm_url.FileName().c_str()));
  f->type=DataPoint::FileInfo::file_type_file;
  if(url->meta_size_available()) {
    f->size=url->meta_size();
    f->size_available=true;
  };
  if(url->meta_created_available()) {
    f->created=url->meta_created();
    f->created_available=true;
  };
  return true;
}

bool DataHandle::check_srm(void) {
  SRM_URL srm_url(c_url.c_str());
  if(!srm_url) return false;
  SRMClient client(srm_url.ContactURL().c_str());
  if(!client) return false;
  odlog(2)<<"check_srm: looking for metadata: "<<c_url.c_str()<<endl;
  unsigned long long int size;
  string checksum;
  if(!client.info(srm_url.FileName().c_str(),srm_url,size,checksum)) return false;
  odlog(1)<<"check_srm: obtained size: "<<size<<endl;
  if(size > 0) url->meta_size(size);
  odlog(1)<<"check_srm: obtained checksum: "<<checksum<<endl;
  if(checksum.length() > 0) url->meta_checksum(checksum.c_str());
  return true;
}


